/*
 * Katobr Route Manager Core
 */
#include <sys/types.h>
#include <sys/resource.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <signal.h>
#include <unistd.h>
#include <fcntl.h>

#include <errno.h>
#include <stdio.h>
#include <string.h>

#define KRM_NEW		0x0
#define KRM_XORED	0x1
#define KRM_OLD		0x2

#define KRM_QUEUE_XOR	0x1

struct krm_node_t {
	struct krm_node_t * next;
	int ip;
	int mask;
	unsigned char ip0;
	unsigned char ip1;
	unsigned char ip2;
	unsigned char ip3;
};

struct krm_queue_t {
	struct krm_node_t * next;
	int old;
	int new;
	int xored;
	int flags;

	int add_fd;
	int del_fd;

	pid_t add_pid;
	pid_t del_pid;
};

int krm_open ( const char * filename )
{
	int fd = open(filename, O_RDONLY);
	if (fd < 0) {
		fprintf(stderr, "KRM Open: Unable to open %s: %s\n",
				filename, strerror(errno));
		fprintf(stderr, "KRM: Assuming it's empty\n");
		fd = open("/dev/null", O_RDONLY);
		if (fd < 0) {
			fprintf(stderr, "KRM: Unable to open /dev/null: %s\n",
					strerror(errno));
			return -1;
		}
	}
	return fd;
}
int krm_pexec ( int * pid, const char * filename, const char * arg )
{
	int fd[2];
	pid_t tmp;
	if (pipe(fd)) {
		fprintf(stderr, "Unable to build pipe: %s\n", strerror(errno));
		return -1;
	}
	tmp = fork();
	if (tmp < 0) {
		fprintf(stderr, "Unable to fork: %s\n", strerror(errno));
		return -1;
	} else if (tmp) {
		if (pid)
			*pid = tmp;
		close(fd[0]);
		return fd[1];
	} else {
		close(fd[1]);
		close(0);
		dup2(fd[0], 0);
		close(fd[0]);
		if (execlp(filename, filename, arg, 0)) {
			fprintf(stderr, "Unable to exec %s: %s\n", filename, strerror(errno));
			exit(1);;
		}
	}
}

int krm_queue_exec ( struct krm_queue_t * queue, struct krm_node_t * node, int flag )
{
	//fprintf(stderr, "KRM: Executing commands\n");
	if (flag == KRM_NEW || flag == KRM_OLD) {
		char buf[24];
		int size = snprintf(buf, 24, "%u.%u.%u.%u/%d\n",
				node->ip0, node->ip1, node->ip2, 
				node->ip3, node->mask);
		//fprintf( stderr, "Prepared command: %s", buf);
		if (flag == KRM_NEW)
			write(queue->add_fd, buf, size);
		else
			write(queue->del_fd, buf, size);
	} else if (flag == KRM_XORED) {
		fprintf( stderr, "KRM Exec: Nothing to do for XORED nodes\n");
	} else {
		fprintf( stderr, "KRM Exec: Unknown flag : %d\n", flag);
	}
}

int krm_queue_hooks( struct krm_queue_t * queue, const char *filename )
{
	if (!queue)
		return -1;
	queue->add_fd = krm_pexec( &queue->add_pid, filename, "add");
	queue->del_fd = krm_pexec( &queue->del_pid, filename, "del");
	if (queue->add_fd > 0 && queue->del_fd > 0)
		return 0;
	fprintf(stderr, "KRM: Failed to build hooks.\n");
	return -1;
}

int krm_queue_wait( struct krm_queue_t * queue )
{
	if (!queue)
		return -1;
	write(queue->add_fd, "stop\n", 5);
	close(queue->add_fd);
	//kill(queue->add_pid, SIGPIPE);
	waitpid( queue->add_pid, 0, 0);
	write(queue->del_fd, "stop\n", 5);
	close(queue->del_fd);
	//kill(queue->del_pid, SIGPIPE);
	waitpid( queue->del_pid, 0, 0);
}

int krm_queue_xor ( struct krm_queue_t * head, struct krm_node_t * node )
{
	struct krm_node_t * ptr = (struct krm_node_t *) head;
	int cnt = 0;
	if (!head || !node)
		return 0;
	/// Large - first
	for (; ptr->next && ptr->next->ip > node->ip; ptr = ptr->next, cnt++);
	head->new++;
	if (!ptr->next || ptr->next->ip > node->ip) {
		// No entry equal to node
		krm_queue_exec( head, node, KRM_NEW );
	} else {
		// Found equal node, drop pointer
		struct krm_node_t * tmp = ptr->next;
		ptr->next = ptr->next->next;
		head->xored++;
	}
	return 0;
}

int krm_queue_put ( struct krm_queue_t * head, struct krm_node_t * node )
{
	struct krm_node_t * ptr = (struct krm_node_t *) head;
	int cnt = 0;
	if (!head || !node)
		return 0;
	if (head->flags & KRM_QUEUE_XOR)
		return krm_queue_xor( head, node );
	/// Large - first
	for (; ptr->next && ptr->next->ip > node->ip; ptr = ptr->next, cnt++);
	/*
	if (ptr->next)
		fprintf(stderr, "Trying to insert into queue[%d]: 0x%08x/0x%08x\n", 
			cnt, ptr->next->ip, node->ip);
	else
		fprintf(stderr, "Trying to insert into queue[%d]: NULL/0x%08x\n", 
			cnt, node->ip);
	*/
	head->old++;
	node->next = ptr->next;
	ptr->next = node;
	return 0;
}

int krm_aton( const char * start, const char ** end )
{
	const char * tmp = start;
	int result = 0;
	if (!start) return 0;
	for ( tmp = start; *tmp >= '0' && *tmp <= '9'; tmp++)
		result = 10 * result + *tmp - '0';
	if (end) *end = tmp;
	return result;
}

int krm_print( struct krm_node_t *node )
{
	fprintf(stderr, "KRM: %d.%d.%d.%d/%d;\n", 
			node->ip0, node->ip1, node->ip2, 
			node->ip3, node->mask);
}

int krm_parse( struct krm_node_t *node, const char * data, int size )
{
	const char * ptr = data;
	int b0 = 0, b1 = 0, b2 = 0, b3 = 0, mask = 0; 
	for (; *ptr && (*ptr < '0' || *ptr > '9'); ptr++); //Find first digit
	b0 = krm_aton(ptr, &ptr);
	if (*ptr != '.') return 0;
	b1 = krm_aton(++ptr, &ptr);
	if (*ptr != '.') return 0;
	b2 = krm_aton(++ptr, &ptr);
	if (*ptr != '.') return 0;
	b3 = krm_aton(++ptr, &ptr);
	if (*ptr == '/') { // Mask found, get it
		mask = krm_aton(++ptr, &ptr);
		if (mask > 32) mask = 32;
	} else { // No mask, guess...
		const char * tmp = ptr - 2;
		if (*tmp == '.' && *++tmp == '0')
			mask = 24;
		else
			mask = 32;
	}
	if (mask == 0) mask = 32;
	if (node) {
		int ip = (b0 << 24) | (b1 << 16) | (b2 << 8) | b3;
		int bitmask = (-1 << (32 - mask));
		node->ip  = ip & bitmask;
		node->ip0 = b0;
		node->ip1 = b1;
		node->ip2 = b2;
		node->ip3 = b3;
		node->mask = mask;
		//krm_print(node);
	}
	return ptr - data;
}

struct krm_node_t * krm_alloc( int size ) 
{
	struct krm_node_t * ptr = 
		(struct krm_node_t *) malloc(size * sizeof(struct krm_node_t));
	struct krm_node_t * tmp = ptr;
	fprintf(stderr, "Trying to alloc %d nodes...\n", size);
	if (!ptr)
		return 0;
	for (size--; size > 0; size--, ++ptr) {
		ptr->next = ptr + 1;
		//fprintf(stderr, "Next: %p\n", ptr->next);
	}
	ptr->next = 0;
	return tmp;
}

int krm_parse_chunk( 
		struct krm_queue_t * queue, 
		struct krm_node_t ** alloc, 
		const char * base, 
		int size )
{
	int offset = 0;
	int result;
	int cnt = 0;
	struct krm_node_t * tmp = *alloc;
	struct krm_node_t * node = 0;
	while (offset < size) {
		if (!tmp) {
			// Avarage size of string data - 14.7 chars
			//fprintf(stderr, "Trying to alloc (%d - %d)/ 14 + 1 = %d nodes...\n", 
			//		size, offset, (size - offset)/14 + 1);
			tmp = krm_alloc ( (size - offset) / 14 + 1 );
			if (!tmp) {
				fprintf(stderr, "Failed to alloc krm_node\n");
				*alloc = 0;
				return -1;
			}
		}
		node = tmp;
		tmp = tmp->next;
		node->next = 0;
		result = krm_parse(node, base + offset, size - offset );
		//fprintf(stderr, "Parsed...\n");
		if (result <= 0) {
			fprintf(stderr, "Failed to parse krm_node: %d\n", result);
			break;
		}
		krm_queue_put(queue, node);
		offset += result;
		cnt++;
	}
	fprintf(stderr, "Parsed %d bytes, %d nodes\n", offset, cnt);
	*alloc = tmp;
	return offset;
}

int main() 
{
	char _buf[1024];
	char *buf = _buf;
	int size = 1024;
	int result = 0;
	int bytes = 0;
	int offset = 0;
	struct krm_queue_t  queue;
	struct krm_node_t * node = 0;
	struct krm_node_t * tmp = 0;
	struct timeval tstart, tstop;
	struct rusage  rstart, rstop;
	double time, cputime;
	const char * old_file = "old.net";
	const char * new_file = "new.net";
	int old_fd = 0, new_fd = 0;
	queue.next  = 0;
	queue.old   = 0;
	queue.new   = 0;
	queue.xored = 0;
	queue.flags = 0;

	old_fd = krm_open(old_file);
	new_fd = krm_open(new_file);
	if (old_fd < 0 || new_fd < 0) {
		fprintf(stderr, "KRM: Failed to open files\n");
		return -1;
	}

	krm_queue_hooks(&queue, "./krm.sh");
	fprintf(stderr, "Hooks build...\n");

	fprintf(stderr, "Read Size: %d\n", 1024 - offset);

	getrusage(RUSAGE_SELF, &rstart);
	gettimeofday(&tstart, 0);

	while ((size = read(old_fd, buf + offset, 1024 - offset)) > 0) {
		//fprintf(stderr, "Size: %d\n", size);
		bytes += size;
		size += offset;
		offset = krm_parse_chunk(&queue, &tmp, buf, size - 64);
		//fprintf(stderr, "Moving %d bytes of data\n", size - offset);
		memmove(buf, buf + offset, size - offset);
		offset = size - offset;
		//fprintf(stderr, "New offset: %d\n", offset);
		//fprintf(stderr, "Read Size: %d\n", 1024 - offset);
	}
	fprintf(stderr, "Failed size: %d\n", size);
	if (offset) {
		fprintf(stderr, "Size: %d\n", offset);
		offset = krm_parse_chunk(&queue, &tmp, buf, offset);
	}

	queue.flags |= KRM_QUEUE_XOR;

	offset = 0;
	size = 1024;
	while ((size = read(new_fd, buf + offset, 1024 - offset)) > 0) {
		//fprintf(stderr, "Size: %d\n", size);
		bytes += size;
		size += offset;
		offset = krm_parse_chunk(&queue, &tmp, buf, size - 64);
		//fprintf(stderr, "Moving %d bytes of data\n", size - offset);
		memmove(buf, buf + offset, size - offset);
		offset = size - offset;
		//fprintf(stderr, "New offset: %d\n", offset);
		//fprintf(stderr, "Read Size: %d\n", 1024 - offset);
	}
	fprintf(stderr, "Failed size: %d\n", size);
	if (offset) {
		fprintf(stderr, "Size: %d\n", offset);
		offset = krm_parse_chunk(&queue, &tmp, buf, offset);
	}

	gettimeofday(&tstop, 0);
	getrusage(RUSAGE_SELF, &rstop);
	time = 1.*(tstop.tv_sec - tstart.tv_sec) + 0.000001*(tstop.tv_usec - tstart.tv_usec);
	cputime = 
		1.*(rstop.ru_utime.tv_sec - rstart.ru_utime.tv_sec) 
		+ 0.000001*(rstop.ru_utime.tv_usec - rstart.ru_utime.tv_usec);
	cputime +=
		1.*(rstop.ru_stime.tv_sec - rstart.ru_stime.tv_sec) 
		+ 0.000001*(rstop.ru_stime.tv_usec - rstart.ru_stime.tv_usec);

	size = queue.old + queue.new;
	fprintf(stderr, "Done: Size: %d [Bytes: %d]. Time: %.6fs. Speed: %.3f ip/ms. CPU: %.2f%%\n",
			size, bytes, time, (1.*size)/time, 100.*cputime/time);
	krm_queue_wait (&queue);
	fprintf(stderr, "Hooks terminated...\n");
	return 0;
}
